## Weiler-Atherton Clipping Applet

## Author: Christopher Andrews
## Date: May 2008


from javax.swing import JPanel, BorderFactory
from java.awt import Font, Color, Dimension, BasicStroke
from java.awt.geom import GeneralPath
from java.awt import BasicStroke


class DrawingPanel(JPanel):
    def __init__(self, controller):
        self.controller = controller
        self.setBorder(BorderFactory.createLineBorder(Color.black))
       
        self.setPreferredSize(Dimension(500,375))   
        self.setBackground(Color.white)
        self.mousePressed = self.addPoint


            
    def addPoint(self, event):
        mode = self.controller.mode
        if mode==self.controller.ENTER_CLIP or mode==self.controller.ENTER_SUBJ:
            x = event.getX()
            y = event.getY()
            if mode==self.controller.ENTER_CLIP:
                poly = 'CLIP'
            else:
                poly = 'SUBJECT'
            if self.lastPoint is None or x != self.lastPoint[0] or y != self.lastPoint[1] or mode != self.lastPoint[2]:
                # filtering out double clicks
                self.lastPoint = (x,y,poly)
                self.controller.canvasClick()
                self.controller.polygons[poly].addPoint(x,y)
                self.repaint()

   
        
    def reset(self):
        """
        This function is called to tell the canvas to revert back to start and
        initialize everything again
        """
        self.resetActions() 
        self.repaint()
        
    def resetActions(self):
        self.points = []
        self.tmpPoints = []
        self.newPolygons = []
        self.done = False
        self.lineData = None
        self.segmentData = None
        self.lastPoint = None
        self.clipPath = None

    def applyActions(self, actions):
        """
        Takes in a list of actions and iterators over it, applying them if
        they are meaningful.
        """
        self.tmpPoints = []
        self.lineData = None
        self.segmentData = None
        for action in actions:
            try:
                method = getattr(self,action[0])
                args = action[1:]
                if callable(method):
                    method(args)
                else:
                    self.controller.console.consoleWrite("Error: Attribute %s is not a method" % action[0])

            except AttributeError:
                # this action isn't meant for this object so ignore it
                pass

## Actions ###   


    def showLines(self, args):
        A, B, C, D = args
        self.lineData = (A,B,C,D)
        
    def showSegments(self, args):
        A, B, C, D = args
        self.segmentData = (A,B,C,D)

    def showPoint(self, args):
        point = args[0]
        if len(args) == 2:
            save = args[1]
        else:
            save = False
        if save:
            self.points.append(point)
            point.visible = True

            if self.controller.mode == '2':
                if self.clipPath is None:
                    self.clipPath = GeneralPath()
                    self.clipPath.moveTo(point.x, point.y)
                else:
                    self.clipPath.lineTo(point.x, point.y)
        else:
            self.tmpPoints.append(point)

    def saveIntersections(self, args):
        """
        We have finished identifying intersections - now we are going to be
        walking the boundaries.
        """
           
        self.points = []

 
    def savePolygon(self, args):
        polygon = args[0]
        self.newPolygons.append(polygon)
        self.clipPath = None


    def finish(self, args):
        self.points = []
        self.actions = ()
        self.done = True
        self.repaint()


## Painting ##


    def drawFullLine(self, g, A, B):
        """
        This function takes in a line segment and draws a complete line
        (as far as the viewing region is concerned at least)
        """
        width = self.getWidth()
        height = self.getHeight()
        denom = B.x - A.x
        if denom == 0:
            # this line is parallel to the y axis
            g.drawLine(int(round(A.x)), 0, int(round(A.x)), height)
        else:
            slope = (B.y - A.y) / float(denom)
            y0 = A.y - (slope * A.x)
            y1 = slope*width + y0
            g.drawLine(0, int(round(y0)), width, int(round(y1)))

    def drawSegments(self, g, args):
        A, B, C, D = args

        g.setColor(Color.green)
        oldStroke = g.getStroke()
        g.setStroke(BasicStroke(4))
        g.drawLine(int(round(A.x)), int(round(A.y)),
                   int(round(B.x)), int(round(B.y)))
        g.drawLine(int(round(C.x)), int(round(C.y)),
                   int(round(D.x)), int(round(D.y)))
        g.setStroke(oldStroke)
        self.drawPoint(g, A.x, A.y, 8, Color.blue)
        self.drawPoint(g, B.x, B.y, 8, Color.blue)
        self.drawPoint(g, C.x, C.y, 8, Color.blue)
        self.drawPoint(g, D.x, D.y, 8, Color.blue)


    def drawLines(self, g, args):
        A, B, C, D = args
          
        self.drawSegments(g, args)
        g.setColor(Color.green)
        oldStroke = g.getStroke()
        g.setStroke(BasicStroke(.5))
        self.drawFullLine(g, A, B)
        self.drawFullLine(g, C, D)
        g.setStroke(oldStroke)
        self.drawPoint(g, A.x, A.y, 8, Color.blue)
        self.drawPoint(g, B.x, B.y, 8, Color.blue)
        self.drawPoint(g, C.x, C.y, 8, Color.blue)
        self.drawPoint(g, D.x, D.y, 8, Color.blue)

    def drawPoint(self, g, x, y, size, color):
        g.setColor(color)
        half = size / 2.
        g.fillOval(int(round(x-half)), int(round(y-half)) , size, size)

    def paintComponent(self, g):
    
        self.super__paintComponent(g)
        mode = self.controller.mode
        if mode == self.controller.ENTER_CLIP:
           self.controller.polygons['CLIP'].draw(g)
        else:
           self.controller.polygons['CLIP'].fill(g)
           if mode == self.controller.ENTER_SUBJ:
               self.controller.polygons['SUBJECT'].draw(g)
           elif not self.done:
               self.controller.polygons['SUBJECT'].fill(g)

        for polygon in self.newPolygons:
            polygon.fill(g)
            

        for point in self.points:
            self.drawPoint(g, point.x, point.y, 8, Color.yellow)

 

        if self.lineData is not None:
            self.drawLines(g, self.lineData)

        if self.segmentData is not None:
            self.drawSegments(g, self.segmentData)

        if self.clipPath is not None:
            g.setColor(Color.yellow)
            origStroke = g.getStroke()
            g.setStroke(BasicStroke(2, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND))
            g.draw(self.clipPath)
            g.setColor(Color.black)
            g.setStroke(origStroke)

            
        for point in self.tmpPoints:
            self.drawPoint(g, point.x, point.y, 8, Color.red)

            

        self.controller.polygons['CLIP'].showLabels(g)
        if not self.done:
            self.controller.polygons['SUBJECT'].showLabels(g)
        else:
            for polygon in self.newPolygons:
                polygon.showLabels(g) 


        g.setFont(Font("SansSerif", Font.BOLD, 14))
        mode = self.controller.mode
        g.drawString('Phase %s: %s' % (mode, self.controller.messages[mode]), 10,20)
