
## Weiler-Atherton Clipping Applet

## Author: Christopher Andrews
## Date: May 2008


from java.awt import Color, Polygon, AlphaComposite
from java.awt import Font, FontMetrics
from jarray import zeros

import Utilities
import math



class CVPolygon(Polygon):
    def __init__(self,c=Color.white, alpha = 1, name=0):
        Polygon.__init__(self)
        self.setColor(c, alpha)
        self.valid = False
        self.name = name
        self.pointList = None
        self.lastPoint = None
  
    def setColor(self, c, alpha):
        self.color = c
        self.alpha = alpha


    def addPoint(self, x,y, name = ''):
        x, y =  int(round(x)), int(round(y))
        if self.lastPoint is not None and x == self.lastPoint[0] and y == self.lastPoint[1]:
            return
        self.lastPoint = (x,y)
        Polygon.addPoint(self, x, y)
        p = PointNode(x,y)
        p.name = name
        p.visible = True
        p.setT(1.0, self.name)
        p.setNext(None, self.name)
        if self.pointList is None:
            self.valid = True
            self.pointList = p
        else:
            plist = self.pointList
            while plist.next[self.name] is not None:
                plist = plist.next[self.name]
            plist.setNext(p, self.name) 


    def points(self):
        pointList = []
        i = self.getPathIterator(None)
        pointData = zeros(6, 'f')
        while not i.isDone():
            segType = i.currentSegment(pointData)
            if segType == 0 or segType == 1:
                pointList.append((int(pointData[0]), int(pointData[1])))
            i.next()
        return pointList


    def reversePoints(self):
        prev = None
        current = self.pointList
       
        while current is not None:
            next = current.next[self.name]
            current.next[self.name] = prev
            prev = current
            current = next
        self.pointList = prev

    def buildName(self, i, names):
        """
        This function builds a unique name for a number given and alphabet
        of possible characters.
        """
        s = [names[i % len(names)]]
        i = (i/len(names)) - 1
        
        while i >= 0:
            s.append(names[i % len(names)])
            i = (i/len(names)) - 1
  
        s.reverse()
        return ''.join(s)

    def validate(self, names):
        """
        This makes sure the polygon is simple, reverses it if necessary
        and applies labels based on the string names.
        """
     
        if not self.isSimple():
            return False

        # reverse if necessary
        if not self.isCCW():
            self.reversePoints()

       # label the points
        point = self.pointList
        i = 0
        while point is not None:
            point.name = self.buildName(i, names)
            point.visible = True
            i += 1
            point = point.next[self.name]

        return True

    

    def printList(self):
        plist = self.pointList
        while plist is not None:
            print plist
            plist = plist.next[self.name]        

    def isCCW(self):
        """
        This function determines if a polygon is specified in clockwise (CW) or
        counterclockwise (CCW) direction, returning True if it is CCW and False
        otherwise.

        The test is performed by performing a cross product of the line
        segments BC x BA, where A,B and C are three sequential points in the
        polygon. As all points lie on the XY-plane, we can ignore
        the X and Y components of the resultant vector (they will be zero). We
        only care about the sign of the Z component - positive means CCW,
        while negative means CW [recall that the Y-axis is upside down].

        Unfortunately, this technique only works if the angle is convex. To
        handle this, we first find the leftmost point (any -most point would do,
        we just need an extreme point to guarantee that it is the apex of a
        convex angle). This will be the middle point.
        """
        points = self.points()
        minIndex = 0

        for i in range(1, len(points)):
            if points[i][0] < points[minIndex][0]:
                minIndex = i

        B = points[minIndex]
        A = points[minIndex - 1]
        C = points[(minIndex + 1) % len(points)]
        

        z = ((A[0] - B[0])*(C[1] - B[1])) - ((A[1] - B[1])*(C[0] - B[0]))
               
        return (z > 0)

    def isSimple(self):
        """
        This function determines if the polygon is a simple one.
        """
        A = self.pointList
        B = A.next[self.name]
        C = B.next[self.name]
        while C is not None:
            while C is not None:
                
                D = C.next[self.name]
                if D is None: D = self.pointList
             
                if A != D:
                    intersection = Utilities.intersection(A,B,C,D)
                    if intersection is not None and intersection[1] is not None:
                        # not parallel and within segment AB
                        (t, u, inbound) = intersection
                        if (0 <= t <= 1) and (0 <= u <= 1):
                            return False
                C = C.next[self.name]
               
            A = B
            B = B.next[self.name]
            if B is not None:
                C = B.next[self.name]
        return True


    def hideIntersections(self):
        """
        This function flips the visibility of all of the intersections off.
        We use this to help with the back button functionality.
        """
        plist = self.pointList
        while plist is not None:
            if plist.isIntersection():
                plist.visible = False
            plist = plist.next[self.name]       


    def showLabels(self, g):
        g.setColor(Color.black)
        if self.pointList is not None:
            g.setFont(Font("SansSerif", Font.BOLD, 14))
            metrics = g.getFontMetrics()
            current = self.pointList
            while current is not None:
                if current.visible:
                    width = ( metrics.stringWidth(current.name) / 2.) 
                    g.drawString(current.name, current.x - width, current.y -8)
                current = current.next[self.name]
                    

    def draw(self, g, showOutline=True):
        if  self.valid:
            g.setColor(Color.black)
            if showOutline:
                g.drawPolygon(self)
           
            current = self.pointList
            while current is not None:
                if current.visible:
                    g.fillOval(int(round(current.x - 2)), int(round(current.y - 2)), 4, 4)
                                
                current = current.next[self.name]


                
    def fill(self, g):
        if self.valid:
            origComposite = g.getComposite()
            g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, self.alpha))
            g.setColor(self.color)
            g.fillPolygon(self)
            g.setComposite(origComposite)
            self.draw(g, False)
        

class Vector:
    def __init__(self, x,y):
        self.x = x
        self.y = y

    def dot(self, v):
        return self.x*v.x + self.y*v.y

    def perp(self):
        return Vector(-self.y, self.x)

    def magnitude(self):
        return math.sqrt(self.x*self.x + self.y*self.y)

    def normalize(self):
        magnitude = self.magnitude()
        self.x /= magnitude
        self.y /= magnitude

    def __add__(self, other):
        if isinstance(other, PointNode):
            return PointNode(self.x + other.x, self.y + other.y)
        elif isinstance(other, Vector):
            return Vector(self.x + other.x, self.y + other.y)
        else:
            return NotImplemented

    def __sub__(self, other):
        if isinstance(other, PointNode):
            return PointNode(self.x - other.x, self.y - other.y)
        elif isinstance(other, Vector):
            return Vector(self.x - other.x, self.y - other.y)
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return Vector(self.x * other, self.y * other)
        else:
            return NotImplemented 

    def __div__(self, other):
        if isinstance(other, int) or isinstance(other, float):
            return Vector(self.x / other, self.y / other)
        else:
            return NotImplemented 

    def __str__(self):
        return "V(%f, %f)" % (self.x, self.y)

    def __repr__(self):
        return "V(%f, %f)" % (self.x, self.y) 


class PointNode:
    def __init__(self, x,y, name=""):
        self.name = name
        self.x = x
        self.y = y
        self.inbound = False
        self.visited = False
        self.visible = False
             
        # links to the next point in the list - this is a dictionary based on
        # the name of the polygon currently being traced. Intersections should
        # have two pointers, normal nodes will have one.
        self.next = {}

        # this tells us the t value of an intersection. Again this is a
        # dictionary as we have different t values for the two polygons
        self.t = {}

    def visit(self):
        self.visited = True

    def setNext(self, pn, polygon):
        """
        Set the next pointer to the PointNode pn. The polygon variable tells us
        which polygon the point belongs to.
        """
        self.next[polygon] = pn

    def setT(self, t, polygon):
        """
        This sets the t value of this point for polygon polygon
        """
        self.t[polygon] = t

        
    def isIntersection(self, name=None):
              
        if name is None:
            return len(self.next) > 1
        else:
            try:
                return self.t[name] < 1
            except KeyError:
                return False
            
    def __sub__(self, other):
        if isinstance(other, PointNode):
            return Vector(self.x - other.x, self.y - other.y)
        elif isinstance(other, Vector):
            return PointNode(self.x - other.x, self.y - other.y)
        else:
            return NotImplemented

    def __add__(self, other):
        if isinstance(other, Vector):
            return PointNode(self.x + other.x, self.y + other.y)
        else:
            return NotImplemented

    def __str__(self):
        return "%s(%f, %f)" % (self.name, self.x, self.y)

    def __repr__(self):
        return "%s(%f, %f)" % (self.name, self.x, self.y)
