## Weiler-Atherton Clipping Applet

## Author: Christopher Andrews
## Date: May 2008


from javax.swing import JPanel, BorderFactory
from java.awt import Color, Dimension, Font,  AlphaComposite
from java.awt.geom import GeneralPath
from java.awt import BasicStroke

class PointListPanel(JPanel):
    def __init__(self, controller):
        self.controller = controller
        self.setBorder(BorderFactory.createLineBorder(Color.black))
        self.setBackground(Color.white)
        self.font = Font("SansSerif", Font.BOLD, 12)
        self.pointWidth = 6
        self.reset()

    def getPreferredSize(self):
        return self.favoredSize

    def reset(self):
        self.resetActions()
        self.favoredSize = Dimension(60,100)
        self.repaint()
        

    def resetActions(self):
        self.currentPoint = None
        self.currentPolygon = ''
        self.visitedPoints = []
        self.pointPositions = {'CLIP':{}, 'SUBJECT':{}}
        self.clipPath = None
        self.oldPaths = []


    def applyActions(self, actions):
        """
        Takes in a list of actions and iterators over it, applying them if
        they are meaningful.
        """
        for action in actions:
            try:
                method = getattr(self,action[0])
                args = action[1:]
                if callable(method):
                    method(args)
                else:
                    self.controller.console.consoleWrite("Error: Attribute %s is not a method" % action[0])

            except AttributeError:
                # this action isn't meant for this object so ignore it
                pass



    def calculatePointPositions(self):
        fontHeight = self.metrics.getHeight()

        for poly in ('CLIP', 'SUBJECT'):
            current = self.controller.polygons[poly].pointList
            if poly == 'CLIP':
                currentPoint = [self.endPoint + 15, self.clipBaseLine]
                y =  currentPoint[1] + 10 + self.pointWidth/2
            else:
                currentPoint = [self.endPoint + 15, self.subjectBaseLine]
                y = currentPoint[1] - 10 - fontHeight + self.pointWidth/2

                
            while current is not None:
                charWidth = self.metrics.stringWidth(current.name)
                x = currentPoint[0] + charWidth / 2
                self.pointPositions[poly][current.name] = (x,y)
                currentPoint[0] += charWidth + 15
                current = current.next[poly]



## Actions ##

    def showPoint(self, args):
        point = args[0]
        if len(args) == 2:
            save = args[1]
        else:
            save = False
        self.currentPoint = point
        if save and self.currentPolygon != '':
            self.visitedPoints.append(point)
            pos = self.pointPositions[self.currentPolygon][point.name]
            if self.clipPath is None:
                self.clipPath = GeneralPath()
                self.clipPath.moveTo(pos[0], pos[1])
            else:
                if (self.lastPoint[1] == pos[1] and self.lastPoint[0] > pos[0]) or abs(self.lastPoint[0] - pos[0]) > 75:
                    # the line is wrapping around - insert a vertical rise for cleaner line
                    if self.currentPolygon == 'CLIP':
                        self.clipPath.lineTo(self.lastPoint[0], self.lastPoint[1] + 15)
                        self.clipPath.lineTo(pos[0], pos[1] + 15)
                    else:
                        self.clipPath.lineTo(self.lastPoint[0], self.lastPoint[1] - 15)
                        self.clipPath.lineTo(pos[0], pos[1] - 15)
                self.clipPath.lineTo(pos[0], pos[1])
            self.lastPoint = pos

    def savePolygon(self, args):
        self.oldPaths.append(self.clipPath)
        self.clipPath = None

    def switchPolygon(self, args):
        self.currentPolygon = args[0]
        pos = self.pointPositions[self.currentPolygon][self.currentPoint.name]
        self.lastPoint = pos
        self.clipPath.lineTo(pos[0], pos[1])
        

    def saveIntersections(self, args):
        """
        This is our signal that the insersecting phase is complete and it is
        time to start clipping
        """
        self.currentPolygon = 'SUBJECT'
        self.calculatePointPositions()
            
## Painting functions ##
        
    def paintHeaders(self, g):
        """
        This function paints the name and the legend boxes on the left side of
        the view

        As a side effect, this calculates three values,
        endPoint - the rightmost point of the header
        clipBaseLine - the horizontal line all clip points should rest on
        subjectbaseLine - the line all subbject points should rest on
        """
        # calculate metrics
        size = self.getSize()
        boxSize = self.metrics.getHeight()
        self.endPoint = self.metrics.stringWidth("SUBJECT :") + 25 + boxSize
        self.clipBaseLine = 25
        self.subjectBaseLine = size.height - 10

        # draw the clip header
        g.drawString("CLIP :", self.endPoint - self.metrics.stringWidth("CLIP :"), self.clipBaseLine) 
        g.setColor(Color.lightGray)
        g.fillRect(15, self.clipBaseLine - boxSize, boxSize, boxSize)
        g.setColor(Color.black)
        g.drawRect(15, self.clipBaseLine - boxSize, boxSize, boxSize)
       

        # draw the subject header
        g.drawString("SUBJECT :", 25 + boxSize, self.subjectBaseLine)
        origComposite = g.getComposite()
        g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, .3))
        g.setColor(Color.blue)
        g.fillRect(15, self.subjectBaseLine - boxSize, boxSize, boxSize)
        g.setComposite(origComposite)
        g.setColor(Color.black)
        g.drawRect(15, self.subjectBaseLine - boxSize, boxSize, boxSize)
    
 

    def paintPoint(self, g, point, x, y, size, polygon):
        if point in self.visitedPoints:
            g.setColor(Color.yellow)
            g.fillOval(x, y, size, size)
        
        if self.currentPolygon == polygon and point == self.currentPoint:
            g.setColor(Color.red)
            g.fillOval(x, y, size, size)

        g.setColor(Color.black)
        g.drawOval(x, y, size, size)
    
    def paintComponent(self, g):

        self.super__paintComponent(g)

        origStroke = g.getStroke()
        g.setStroke(BasicStroke(2, BasicStroke.CAP_BUTT, BasicStroke.JOIN_ROUND))
        g.setColor(Color.lightGray)
        for path in self.oldPaths:
            g.draw(path)

        g.setColor(Color.black)
        if self.clipPath is not None:
            g.draw(self.clipPath)

        g.setStroke(origStroke)
        g.setFont(self.font)

        self.metrics = g.getFontMetrics()
        self.paintHeaders(g)
        fontHeight = self.metrics.getHeight()
         
        for poly in ('CLIP', 'SUBJECT'):
            current = self.controller.polygons[poly].pointList
            if poly == 'CLIP':
                currentPoint = [self.endPoint + 15, self.clipBaseLine]
                y =  currentPoint[1] + 10
            else:
                currentPoint = [self.endPoint + 15, self.subjectBaseLine]
                y = currentPoint[1] - 10 - fontHeight

            while current is not None:
                if current.visible:
                    charWidth = self.metrics.stringWidth(current.name)
                    g.drawString(current.name,currentPoint[0], currentPoint[1])

                    x = currentPoint[0] + (charWidth - self.pointWidth) / 2
                    self.paintPoint(g, current, x, y, self.pointWidth, poly)
                    currentPoint[0] += charWidth + 15

                current = current.next[poly]
            if currentPoint[0] > self.favoredSize.width:
                self.favoredSize.width = currentPoint[0]
                self.revalidate()

        


        ##self.currentPoint = None
